#include "arm9.h"

#include "wifi.h"

#ifdef MKDS_COURSE_HACK
#include "mkdshack.h"
#endif

#ifdef SCROLL_TEXT
#include "scroller.h"
#endif


typedef struct {
	int id;
	int id2;
	u32 CRCs;
	int invalid;
	u8 _10[8];
} PACKED CARDINFO;
static CARDINFO *cardinfo=(CARDINFO*)0x27ff800;

u8 gba_logo[] = {
	0x24, 0xFF, 0xAE, 0x51, 0x69, 0x9A, 0xA2, 0x21, 
	0x3D, 0x84, 0x82, 0x0A, 0x84, 0xE4, 0x09, 0xAD, 
  0x11, 0x24, 0x8B, 0x98, 0xC0, 0x81, 0x7F, 0x21, 
  0xA3, 0x52, 0xBE, 0x19, 0x93, 0x09, 0xCE, 0x20, 
  0x10, 0x46, 0x4A, 0x4A, 0xF8, 0x27, 0x31, 0xEC, 
  0x58, 0xC7, 0xE8, 0x33, 0x82, 0xE3, 0xCE, 0xBF, 
  0x85, 0xF4, 0xDF, 0x94, 0xCE, 0x4B, 0x09, 0xC1, 
  0x94, 0x56, 0x8A, 0xC0, 0x13, 0x72, 0xA7, 0xFC, 
  0x9F, 0x84, 0x4D, 0x73, 0xA3, 0xCA, 0x9A, 0x61, 
  0x58, 0x97, 0xA3, 0x27, 0xFC, 0x03, 0x98, 0x76, 
  0x23, 0x1D, 0xC7, 0x61, 0x03, 0x04, 0xAE, 0x56, 
  0xBF, 0x38, 0x84, 0x00, 0x40, 0xA7, 0x0E, 0xFD, 
  0xFF, 0x52, 0xFE, 0x03, 0x6F, 0x95, 0x30, 0xF1, 
  0x97, 0xFB, 0xC0, 0x85, 0x60, 0xD6, 0x80, 0x25, 
  0xA9, 0x63, 0xBE, 0x03, 0x01, 0x4E, 0x38, 0xE2, 
  0xF9, 0xA2, 0x34, 0xFF, 0xBB, 0x3E, 0x03, 0x44, 
  0x78, 0x00, 0x90, 0xCB, 0x88, 0x11, 0x3A, 0x94, 
  0x65, 0xC0, 0x7C, 0x63, 0x87, 0xF0, 0x3C, 0xAF, 
  0xD6, 0x25, 0xE4, 0x8B, 0x38, 0x0A, 0xAC, 0x72, 
  0x21, 0xD4, 0xF8, 0x07
};
u8 cart_data[256];

int isValidDSCart() {
	int i;
	u8 game_name[] = "DSBooter";
	u8 game_code[] = "PASS";
	
	swiFastCopy((u32*)0x8000000,(u32*)cart_data,256>>2);
	
	for(i=0;i<8;i++)
		if(cart_data[0xA0 + i] != game_name[i])
			break;
	if(i==8)
		return 1; //Game name passme catcher
	
	for(i=0;i<4;i++)
		if(cart_data[0xAC + i] != game_code[i])
			break;
	if(i==4)
		return 1; //Game code passme catcher
		
	return 0;
}
	

int isValidGBACart() {
	int i,j;
	
	swiFastCopy((u32*)0x8000000,(u32*)cart_data,256>>2);

	for(i=0;i<0x9C;i++)
	{
		if(i==0x98)
		{
			if((cart_data[i+4]&0x7B)!=0x21)
				break;
		}
		else if(i==0x9A)
		{
			if((cart_data[i+4]&0xFC)!=0xF8)
				break;
		}
		else
		{
			if(cart_data[i+4] != gba_logo[i])
				break;
		}
	}
	if(i==0x9C)	//GBA logo is valid, now check the complement.
	{
		for(i=0xA0,j=0;i<=0xBC;i++)
		{
			j-=cart_data[i];
		}
		j-=0x19;
		j&=0xFF;
		if(cart_data[0xBD]==j)
			return 1;
	}
	
	return 0;
}


void Write_PMemory(u32 mem_loc, unsigned char new_value){
  unsigned char * FV_Reg = (unsigned char*)mem_loc;
  * FV_Reg = new_value;
}

unsigned char Read_PMemory(u32 mem_loc){
  unsigned char * FV_Reg = (unsigned char*)mem_loc;
  return (* FV_Reg);
}

void Write_PMemory_16(u32 mem_loc, u16 new_value){
	u16 * FV_Reg = (u16*)mem_loc;
	* FV_Reg = new_value;
}

u16 Read_PMemory_16(u32 mem_loc){
	u16 * FV_Reg = (u16*)mem_loc;
	return (* FV_Reg);
}

void Write_PMemory_32(u32 mem_loc, u32 new_value){
	u32 * FV_Reg = (u32*)mem_loc;
	* FV_Reg = new_value;
}

u32 Read_PMemory_32(u32 mem_loc){
	u32 * FV_Reg = (u32*)mem_loc;
	return (* FV_Reg);
}


void interrupthandler() {
	
	
	u32 flags=REG_IF&REG_IE;
	VBLANK_INTR_WAIT_FLAGS|=flags;
	scanKeys();
	
	#ifdef SCROLL_TEXT
		scroll_text();
	#endif
	
	REG_IF=flags;
}

inline void clearmem(u32 dest, u32 size) {
	int i=0;
	swiFastCopy(&i, (u32*)dest, 0x01000000 | (size>>2));
}

inline void copy(void *src, void *dst, int size) {
	swiFastCopy(src,dst,size>>2);
}

char *day_of_week[] = {
	"SUN",
	"MON",
	"TUE",
	"WED",
	"THU",
	"FRI",
	"SAT",
};

void display_info()
{
	char str[33];
	int i;
	
	i=IPC->curtime[5];
	sprintf(str,"           %s %.2d/%.2d/%.2d %.2d:%.2d:%.2d",day_of_week[IPC->curtime[4]],IPC->curtime[2],
		IPC->curtime[3],IPC->curtime[1],i,IPC->curtime[6],IPC->curtime[7]);
	for(i=0;i<10;i++)
	{
		if((PersonalData->name[i] >= 0x20) && (PersonalData->name[i] <= 0x7E))
			str[i] = PersonalData->name[i];
		else
			break;
	}
	printtext(20,str);
	for(i=0;i<32;i++)
		str[i]=' ';
	for(i=0;i<26;i++)
	{
		if((PersonalData->message[i] >= 0x20) && (PersonalData->message[i] <= 0x7E))
			str[i+4] = PersonalData->message[i];
		else
		{
			break;
		}
	}
	printtext(21,str);
	
}


void gfx_init() {
	
	//	vramSetMainBanks(VRAM_A_MAIN_BG_0x6000000, VRAM_B_MAIN_BG_0x6020000, // VRAM_B_LCD
	//		VRAM_C_SUB_BG , VRAM_D_LCD);
	
	POWER_CR=POWER_SWAP_LCDS | POWER_ALL_2D;
	vramSetMainBanks(VRAM_A_MAIN_BG,VRAM_B_MAIN_BG,VRAM_C_SUB_BG,VRAM_D_LCD);
	SUB_DISPLAY_CR=MODE_0_2D|DISPLAY_BG0_ACTIVE;
	DISPLAY_CR=MODE_0_2D;
	SUB_BG0_CR=BG_TILE_BASE(1);
	SUB_BG0_X0=0;
	SUB_BG0_Y0=0;
	swiFastCopy(&font_bin,(u16*)(CHAR_BASE_BLOCK_SUB(1)+32*32),font_bin_size/4);
	swiFastCopy(&fontpal_bin,(u16*)BG_PALETTE_SUB, fontpal_bin_size/4);
	lcdSwap();
}

void hardware_init() {

	REG_IME=0;
	REG_IE=0;
	DISP_SR=0;
	DISPLAY_CR=0;
	SUB_DISPLAY_CR=0;

	//reset DMA
	clearmem(0x40000B0, 0x30);

	//clear VRAM
	POWER_CR=1;
	VRAM_CR=0x80808080;
	VRAM_E_CR=0x80;
	VRAM_F_CR=0x80;
	VRAM_G_CR=0x80;
	VRAM_H_CR=0x80;
	VRAM_I_CR=0x80;
	clearmem(0x6800000,0xA4000);
	VRAM_CR=0;
	VRAM_E_CR=0;
	VRAM_F_CR=0;
	VRAM_G_CR=0;
	VRAM_H_CR=0;
	VRAM_I_CR=0;

	POWER_CR=POWER_SWAP_LCDS | POWER_ALL_2D;
	clearmem(0x5000000, 0x800);	//palette
	clearmem(0x7000000, 0x800);	//OAM

	//swiSetHaltCR(1);
}

void arm9sleep() {
	IPC_ARM9=IPC_SLEEP;
	while(IPC_XKEYS & 0x80)
		swiWaitForVBlank();
}

u32 keyhit() {
	static u32 oldkey=~0;
	u32 k,x;
	k=(~REG_KEYINPUT)&0x3ff;
	x=(oldkey^k)&k;
	oldkey=k;
	return x;
}

void Boot_NDS()
{
	if(!check_card_inserted())
		return;
		
	REG_IME=IME_DISABLE;
	sysSetBusOwners(BUS_OWNER_ARM7,BUS_OWNER_ARM7,BUS_OWNER_ARM7);
	swiDelay(0x200000);
	IPC_ARM9=IPC_NDS;
	while((IPC_ARM7!=IPC_NDS_GO)&&(IPC_ARM7!=IPC_NDS_STOP))
		swiDelay(0);
	if(IPC_ARM7==IPC_NDS_GO)
	{
		hardware_init();
		(*(call0*)0x27ffe24)();
	}
	sysSetBusOwners(BUS_OWNER_ARM9,BUS_OWNER_ARM9,BUS_OWNER_ARM9);
	swiDelay(0x200000);
	REG_IME=IME_ENABLE;
}

void Boot_PASSME()
{
	sysSetBusOwners(BUS_OWNER_ARM7,BUS_OWNER_ARM7,BUS_OWNER_ARM7);
	swiDelay(0x200000);
	hardware_init();
	IPC_ARM9=IPC_PASSME;
	*(u32*)0x27ffe24=0x027ffe28;
	*(u32*)0x27ffe28=0xe51ff00c;
	((call0)0x27ffe28)();
}

void Boot_GBA()
{
		sysSetBusOwners(BUS_OWNER_ARM7,BUS_OWNER_ARM7,BUS_OWNER_ARM7);
		swiDelay(0x200000);
		hardware_init();
		if(PersonalData->gbaScreen)
			POWER_CR=1;
		else
			POWER_CR=POWER_SWAP_LCDS|1;
		IPC_ARM9=IPC_GBA;
}

void AutoStart(bool startpressed)
{
	int i,j;
	
	swiWaitForVBlank();
	
	//i=KEY_START|KEY_SELECT;
	i=keysHeld();
	
	i&=(KEY_START|KEY_SELECT|KEY_A|KEY_B|KEY_X|KEY_Y|KEY_R|KEY_L); //Filter off the keys used here.
	
	//Force bootup regardless of autostart flag/settings.
	if((i==(KEY_R|KEY_L|KEY_A)) && (!cardinfo->invalid))	//R + L + A forces NDS
		Boot_NDS();
		
	if((i==(KEY_R|KEY_L|KEY_Y))&&isValidGBACart()) //R + L + Y forces GBA boot
		Boot_GBA();
		
	if((i==(KEY_R|KEY_L|KEY_X))) //R + L + X forces PASSME boot
		Boot_PASSME();
	
	if(i==(KEY_START|KEY_SELECT)) //If Start & select is pressed, do not autostart, period
		return;	
		
		
	
	
	j=Read_PMemory(0x027FFCE4);
	j=(j>>6)&1;
#ifdef USE_AUTOSTART_FLAG
	if(j==0)
	{
		if(!(cardinfo->invalid)) { //Check to see if the Autostart flag in NDS card is set.
			j=Read_PMemory(0x027FFE1F);  //If it is, the DS card will boot here and now.
			j=(j>>2)&1;									//Unlike official firmware, autostart can be skipped.
		}
	}
#endif	
	if((j==1) || startpressed)
	{
		if(!(cardinfo->invalid)) { //Boot DS game card
			Boot_NDS();
		}
		
		
		
		else if (isValidDSCart()) { //Boot DS.gba
			if((i==(KEY_A | KEY_B | KEY_X | KEY_Y))&&isValidGBACart())
				Boot_GBA();
			else
				Boot_PASSME();
		}
		else if (isValidGBACart()) { //Boot GBA
			if(i==(KEY_A | KEY_B | KEY_X | KEY_Y)) //Boot GBA as DS.gba
				Boot_PASSME();
			else //Boot GBA as GBA
				Boot_GBA();
		}
		else
			if(i==(KEY_A | KEY_B | KEY_X | KEY_Y))
				Boot_PASSME(); //If it gets this far, the user definitely has problems
											 //Properly formatting his flashcart, to work in earlier passes, and at least detect
											 //as a valid GBA cart.
	}
}


void clear_screen() {
	int line=0;
	u16 *p=(u16*)(SCREEN_BASE_BLOCK_SUB(0)+line*64);
	int i,j;
	for(i=0;i<24;i++)
		for(j=0;j<32;j++)
			*p++ = ' ';
}

enum { //Keep the Following enums in the same order as the menu text.
			 //This allows for menu items to be defined as compile in options.
	MAINMENU_0,
	MAINMENU_LOADDSCARD,
	MAINMENU_LOADPASSME,
	MAINMENU_DSDOWNLOAD,
	MAINMENU_LOADGBA,
	
	MAINMENU_CONFIG,
	MAINMENU_ALARMCLOCK,
	
	#ifdef MEMORY_DUMP
		MAINMENU_MEMDUMP,
	#endif
	
	#ifdef PROGRAMMING_MODE
		MAINMENU_PROGRAMMING,
	#endif
	
	#ifdef MKDS_COURSE_HACK
		MAINMENU_MKDSHACK,
	#endif
	
	#ifdef DS_CARD_DUMP
		MAINMENU_DUMPDS,
	#endif
	
	MAINMENU_POWEROFF,
	

} MAINMENU_SELECTED;
char *text[]={
 "== FWNITRO " FWNITRO_VERSION " ==",
 "LOAD DS GAME CARD",
 "LOAD DS FROM GBA",
 "DS DOWNLOAD PLAY",
 "LOAD GBA   ",
 
 "CONFIGURATION",
 "ALARM CLOCK",
 
 #ifdef MEMORY_DUMP
 "MEMDUMP",
 #endif
 
 #ifdef PROGRAMMING_MODE
 "PROGRAMMING MODE",
 #endif
 
 #ifdef MKDS_COURSE_HACK
 "MKDS OFFLINE COURSE HACK",
 #endif
 
 #ifdef DS_CARD_DUMP
 "DUMP DS CARD",
 #endif
 
 "POWER OFF",

 
 };
 
char *textdis[]={
 "NO DS GAME CARD  ",
 "NO GBA CART"};
 
void printtext(int line, char *str)
{
	char *s=str;
	u16 *p=(u16*)(SCREEN_BASE_BLOCK_SUB(0)+line*64);
	while (*s) *p++=*s++;
} 
 
void printtext2(int line, char *str)
{
	char *s=str;
	u16 *p=(u16*)(SCREEN_BASE_BLOCK_SUB(0)+line*64);
	while (*s) *p++=(*s++)|0x1000;
	*p++=0x107f;
} 
 
 
#define OPTS (sizeof(text)/sizeof(char*))
void option(int line, int select) {
	char *s;
	u16 *p=(u16*)(SCREEN_BASE_BLOCK_SUB(0)+line*64);
	if(line==MAINMENU_LOADDSCARD)
	{
		if(cardinfo->invalid)
			s=textdis[0];
		else
			s=text[line];
	}
	else if (line==MAINMENU_LOADGBA)
	{
		if(isValidGBACart())
			s=text[line];
		else
			s=textdis[1];
	}
	else
		s=text[line];
	*p++ = select? (0x1000|'*') : ' ';
	while(*s) *p++=(*s++)|(select?0x1000:0);
}

u32 curpal1;

int main() {
	u32 i,selected;
	u32 k;
	
	#ifdef SCROLL_TEXT
	scroll_pos=25;
	scroll_update=0;
	scroll_enabled=false;
	#endif

	sysSetBusOwners(BUS_OWNER_ARM9,BUS_OWNER_ARM9,BUS_OWNER_ARM9);
	swiDelay(0x200000);		//bigger delay reduces sound pop
	

	hardware_init();	
	gfx_init();

	DISP_SR=DISP_VBLANK_IRQ;
	IRQ_HANDLER=interrupthandler;
	REG_IE=IRQ_VBLANK;
	REG_IF=~0;
	REG_IME=IME_ENABLE;
	
	AutoStart(false);
	
	initKeyboard();


	IPC_ARM7=-1;
	selected=0;
	curpal1=Read_PMemory(0x027FFC82);
	set_cur_palette();
	while(1) {
		#ifdef SCROLL_TEXT
			scroll_enabled=true;
		#endif
		swiWaitForVBlank();
		display_info();
		
		for(i=0;i<OPTS;i++)
			option(i,i==selected);

		k=keysDown();
		if(IPC_XKEYS & 0x80)	//lid closed
			arm9sleep();
		if(k&KEY_UP)
			selected--;
		if(k&KEY_DOWN)
			selected++;
		//if(k&KEY_L)
		//	lcdSwap();
		if(k&KEY_L)
			IPC_ARM9=IPC_SET_TOP_LIGHT;
		if(k&KEY_R)
			IPC_ARM9=IPC_SET_BOTTOM_LIGHT;
		if(k&KEY_START)
			AutoStart(true);
		selected%=OPTS;
		if(selected==0 && (k&KEY_UP)) selected=OPTS-1;
		if(selected==0 && !(k&KEY_UP)) selected=1;
		if(k&KEY_SELECT) {
			switch(selected) {
				
				#ifdef DS_CARD_DUMP
				case MAINMENU_DUMPDS:
					#ifdef SCROLL_TEXT
						scroll_enabled=false;
					#endif
					clear_screen();
					Dump_Card();
					clear_screen();
					break;
					#endif
				
				#ifdef MKDS_COURSE_HACK
				case MAINMENU_MKDSHACK:   //MKDS Offline Course Hack
					#ifdef SCROLL_TEXT
						scroll_enabled=false;
					#endif
					clear_screen();
					menu_mkds_hack();
					clear_screen();
					break;
				#endif
					
				
				case MAINMENU_LOADDSCARD:		//NDS
					Boot_NDS();
					break;

				case MAINMENU_LOADPASSME:		//passme
					Boot_PASSME();
					break;

				case MAINMENU_LOADGBA:		//GBA
					if(isValidGBACart())
					{
						Boot_GBA();
					}
					break;

			#ifdef MEMORY_DUMP
				case MAINMENU_MEMDUMP:		//memdump
					#ifdef SCROLL_TEXT
						scroll_enabled=false;
					#endif
					clear_screen();
					memdump();
					clear_screen();
					break;
			#endif

			#ifdef PROGRAMMING_MODE
				case MAINMENU_PROGRAMMING:		//progmode
					clear_screen();
					menu_progmode();
					clear_screen();
					break;
			#endif

				case MAINMENU_CONFIG:		//config
					clear_screen();
					menu_config();
					clear_screen();
					break;

				case MAINMENU_ALARMCLOCK:
					clear_screen();
					alarmclock();
					clear_screen();
					break;

				case MAINMENU_POWEROFF:		//off
					IPC_ARM9=IPC_POWEROFF;
					break;
			}
		}
	}
	return 0;
}

enum {
	MENUCONGIF_0,
	MENUCONFIG_WIFI_1,
	MENUCONFIG_WIFI_2,
	MENUCONFIG_WIFI_3,
	MENUCONFIG_USERPREF,
	MENUCONFIG_RTC,
#ifdef PROTECTED_ENABLED
	MENUCONFIG_PROTECTED,
#endif
	MENUCONFIG_RETURN,
} MENUCONFIG_SELECTED;
char *text2[]={
 "== CONFIGURATION ==",
 "WIFI 1",
 "WIFI 2",
 "WIFI 3",
 "USER PREF",
 "REAL TIME CLOCK",
#ifdef PROTECTED_ENABLED
 "PROTECTED",
#endif
 "RETURN TO MAIN"};
#define OPTS2 (sizeof(text2)/sizeof(char*))

void option2(int line, int select) {
	u16 *p=(u16*)(SCREEN_BASE_BLOCK_SUB(0)+line*64);
	char *s=text2[line];
	*p++ = select? (0x1000|'*') : ' ';
	while(*s) *p++=(*s++)|(select?0x1000:0);
}

void menu_config() {
	u32 i,selected;
	u32 k;

	selected=0;
	while(1) {
		swiWaitForVBlank();
		
		for(i=0;i<OPTS2;i++)
			option2(i,i==selected);

		k=keysDown();
		if(IPC_XKEYS & 0x80)	//lid closed
			arm9sleep();
		if(k&KEY_UP)
			selected--;
		if(k&KEY_DOWN)
			selected++;
		if(k&KEY_START)
			return;
		selected%=OPTS2;
		if(selected==0 && (k&KEY_UP)) selected=OPTS2-1;
		if(selected==0 && !(k&KEY_UP)) selected=1;
		if(k&KEY_SELECT) {
			switch(selected) {

				case MENUCONFIG_WIFI_1:		//Wifi 1
				case MENUCONFIG_WIFI_2:		//Wifi 2
				case MENUCONFIG_WIFI_3:		//Wifi 3
					clear_screen();
					menu_wifi(selected - MENUCONFIG_WIFI_1 + 1);
					clear_screen();
					break;

				case MENUCONFIG_USERPREF:		//User pref
					clear_screen();
					menu_userpref();
					clear_screen();
					break;
					
				case MENUCONFIG_RTC:
					clear_screen();
					menu_rtcsetting();
					clear_screen();
					break;
					
			#ifdef PROTECTED_ENABLED
				case MENUCONFIG_PROTECTED:
					break;
			#endif
			
			}
			if(selected==MENUCONFIG_RETURN)
				break;
		}
	}
}

void option3(char * txt[], int line, int select) {
	char *s;
	u16 *p=(u16*)(SCREEN_BASE_BLOCK_SUB(0)+line*64);
	s=txt[line];
	*p++ = select? (0x1000|'*') : ' ';
	while(*s) *p++=(*s++)|(select?0x1000:0);
}






